# ExceptHook by David Bowland
# ./addons/eventscripts/excepthook/excepthook.py

"""
Catches errors raised by EventScripts Python addons and outputs them to text files. Those error messages can then be passed on to addon authors for correction. Requires only ES 2.0+
"""


import es
import os.path
import sys
import time
import traceback

from collections import deque

import psyco
psyco.full()


info = es.AddonInfo()
info.name     = 'ExceptHook'
info.version  = '2'
info.basename = 'excepthook'
info.url      = 'http://addons.eventscripts.com/addons/view/' + info.basename
info.author   = 'SuperDave'


###


class ExceptionLog(object):
   filepath = es.getAddonPath('excepthook') + '/exceptions.txt'
   addondir = str(es.ServerVar('eventscripts_addondir')).replace('\\', '/')

   def __init__(self):
      self.lasthook  = sys.excepthook
      sys.excepthook = self.excepter

   def excepter(self, *a):
      try:
         self.logError(traceback.format_exception(*a))
      except:
         self.lasthook(*sys.exc_info())
      self.lasthook(*a)

   ###

   def logError(self, lines):
      # Split lines by new line characters
      lines = filter(bool, '\n'.join(lines).split('\n'))

      # Find end of current traceback omitting exception
      traceback_end = -1
      for count, excp_line in enumerate(lines):
         split_line = excp_line.split('"')
         if len(split_line) == 3 and split_line[0].strip() == 'File':
            traceback_end = count

      if not traceback_end:
         raise IndexError, 'Could not find file traceback originated from'

      """ Find path to addon for writing traceback to file """
      addonpath = lines[traceback_end].split('"')[1].replace('\\', '/')
      # Last file is always followed by the error line unless the non-compiled source is missing
      if os.path.isfile(addonpath): 
         traceback_end += 1

      # Write exception to main log file
      currenttime = time.strftime('%x %X')
      addon = addonpath.replace(self.addondir, '')
      f = open(self.filepath, 'a')
      f.write(currenttime + ' file > ' + addon[1:] + '\n\n')
      f.close()

      """ Write exception to addon-specific file """
      # Read exceptions already in file
      filename   = addonpath.rsplit('.', 1)[0] + '_errors.txt'
      exceptions = deque()
      if os.path.isfile(filename):
         excp = []
         f = open(filename)
         for fileline in f.readlines():
            if fileline.strip():
               excp.append(fileline)
            else:
               exceptions.append(excp)
               excp = []
         f.close()

      """ Search log for the exception and insert the exception at the correct position """
      excp_count = []
      found_exception = -1
      for excp_count, excp in enumerate(exceptions):
         if len(excp) > traceback_end:
            found_exception = excp_count
            for count, excp_line in enumerate(excp):
               if (excp_line.split('"')[0].strip() == 'File' and count > traceback_end) or (count <= traceback_end and excp_line.strip('\n') != lines[count]):
                  found_exception = -1
                  break
            if found_exception >= 0:
               break

      # If we found the exceptions look for the same error
      if found_exception >= 0:
         error = lines[traceback_end + 1:]
         excp  = exceptions[found_exception]
         excp_errors = excp[traceback_end + 1:]

         error_index = 0
         excp_count  = -1
         for count, excp_line in enumerate(excp_errors):
            if excp_line.strip('\n') == error[error_index]:
               if not error_index:
                  excp_count = count
               error_index += 1
               if error_index == len(error):
                  break
            else:
               error_index = 0
               excp_count  = -1

         # If we found the same error add the current timestamp to the next line
         if excp_count >= 0:
            if error_index + excp_count >= len(excp):
               excp.append(currenttime)
            else:
               x = traceback_end + error_index + excp_count + 1
               excp[x] = excp[x].strip('\n') + ' * ' + currenttime + '\n'

         # If we didn't find the same error add the error to the end of the existing exception
         else:
            for excp_line in error:
               excp.append(excp_line + '\n')
            excp.append(currenttime + '\n')

      # We didn't find the exception so append it to the end of the file
      else:
         exceptions.append([x + '\n' for x in lines] + [currenttime + '\n'])

      # Write the finished exception log
      write_lines = []
      for excp in exceptions:
         for x in excp:
            write_lines.append(x)
         write_lines.append('\n')

      f = open(filename, 'w')
      f.writelines(write_lines)
      f.close()

   ###

   def unHook(self):
      sys.excepthook = self.lasthook

exceptionlog = ExceptionLog()

###

def unload():
   exceptionlog.unHook()